from __future__ import annotations

from collections import Counter
from dataclasses import dataclass, field
from enum import auto
from functools import cached_property
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from pytest import Config, Item

from pytest_smoke.compat import StrEnum


class SmokeEnvVar(StrEnum):
    SMOKE_TEST_SESSION_UUID = auto()


class SmokeScope(StrEnum):
    FUNCTION = auto()
    CLASS = auto()
    AUTO = auto()
    FILE = auto()
    DIRECTORY = auto()
    ALL = auto()


class SmokeSelectMode(StrEnum):
    FIRST = auto()
    LAST = auto()
    RANDOM = auto()


class SmokeIniOption(StrEnum):
    SMOKE_DEFAULT_N = auto()
    SMOKE_DEFAULT_SCOPE = auto()
    SMOKE_DEFAULT_SELECT_MODE = auto()
    SMOKE_DEFAULT_XDIST_DIST_BY_SCOPE = auto()
    SMOKE_MARKED_TESTS_AS_CRITICAL = auto()


class SmokeDefaultN(int): ...


class SmokeOption:
    def __init__(self, config: Config):
        self.config = config

    @cached_property
    def n(self) -> Optional[int, str]:
        if n := self.config.option.smoke:
            if isinstance(n, SmokeDefaultN):
                # N was not explicitly provided to the option. Apply the INI config value or the plugin default
                n = self._parse_ini_option(SmokeIniOption.SMOKE_DEFAULT_N)
            return n

    @cached_property
    def scope(self) -> str:
        scope = self.config.option.smoke_scope or self._parse_ini_option(SmokeIniOption.SMOKE_DEFAULT_SCOPE)
        assert scope
        return scope

    @cached_property
    def select_mode(self) -> str:
        mode = self.config.option.smoke_select_mode or self._parse_ini_option(SmokeIniOption.SMOKE_DEFAULT_SELECT_MODE)
        assert mode
        return mode

    @cached_property
    def is_scale(self) -> bool:
        return isinstance(self.n, str) and self.n.endswith("%")

    def _parse_ini_option(self, option: SmokeIniOption):
        from pytest_smoke.utils import parse_ini_option

        return parse_ini_option(self.config, option)


class SmokeMarker:
    def __init__(self, *args, mustpass: bool = False, runif: bool = True, **kwargs):
        self.mustpass = bool(mustpass)
        self.runif = bool(runif)

    @classmethod
    def from_item(cls, item: Item) -> Optional[SmokeMarker]:
        if marker := item.get_closest_marker("smoke"):
            return cls(*marker.args, **marker.kwargs)


@dataclass
class MustpassCounter:
    selected: set[Item] = field(default_factory=set)
    failed: set[Item] = field(default_factory=set)


@dataclass
class SmokeCounter:
    collected: Counter = field(default_factory=Counter)
    selected: Counter = field(default_factory=Counter)
    mustpass: MustpassCounter = field(default_factory=MustpassCounter)



from __future__ import annotations

import ast
from collections.abc import Callable
from contextlib import contextmanager, nullcontext
from dataclasses import dataclass, field
from functools import wraps
from itertools import chain
from typing import Optional

import pytest

from pytest_smoke import smoke
from pytest_smoke.plugin import DEFAULT_N, SmokeScope
from pytest_smoke.utils import scale_down

TEST_NAME_BASE = "test_something"
PARAMETRIZED_ARG_NAME_CLS = "p_c"
PARAMETRIZED_ARG_NAME_FUNC = "p_f"

requires_xdist = pytest.mark.skipif(not smoke.is_xdist_installed, reason="pytest-xdist is required")


@dataclass
class TestFileSpec:
    """Specification of a test file"""

    __test__ = False

    test_specs: list[TestFuncSpec | TestClassSpec] = field(default_factory=list)
    test_dir: Optional[str] = None


@dataclass
class TestClassSpec:
    """Specification of a test class"""

    __test__ = False

    name: str
    test_func_specs: list[TestFuncSpec] = field(default_factory=list)
    num_params: int = 0
    param_marker: Callable[[int], str | None] = None

    def __post_init__(self):
        for test_func_spec in self.test_func_specs:
            test_func_spec.test_class_spec = self


@dataclass
class TestFuncSpec:
    """Specification of a test function"""

    __test__ = False

    num_params: int = 0
    test_class_spec: TestClassSpec = field(init=False, default=None)
    param_marker: Callable[[int], str | None] = None
    func_body: str = None


def generate_test_code(test_spec: TestFileSpec | TestFuncSpec) -> str:
    """Generate test code from the test file spec

    :param test_spec: Test file spec or Test func spec
    """
    code = "import pytest\n"

    def add_parametrize_marker(
        *param_names: str,
        num_params: int = 0,
        with_indent: bool = False,
        param_marker: Callable[[int], str | None] = None,
    ):
        nonlocal code
        params = ", ".join(param_names)
        if with_indent:
            code += "\t"
        if param_marker:
            param_values = ", ".join(
                [
                    (f"pytest.param({p}, marks=pytest.mark.{param_marker(p)})" if param_marker(p) else str(p))
                    for p in range(num_params)
                ]
            )
            code += f"@pytest.mark.parametrize('{params}', [{param_values}])\n"
        else:
            code += f"@pytest.mark.parametrize('{params}', range({num_params}))\n"

    def add_class(test_class_spec: TestClassSpec):
        nonlocal code
        if test_class_spec.num_params:
            add_parametrize_marker(
                PARAMETRIZED_ARG_NAME_CLS,
                num_params=test_class_spec.num_params,
                param_marker=test_class_spec.param_marker,
            )
        code += f"class {test_class_spec.name}:\n"

    def add_function(name: str, test_func_spec: TestFuncSpec, test_class_spec: TestClassSpec = None):
        nonlocal code
        params = []
        if test_class_spec and test_class_spec.num_params:
            params.append(PARAMETRIZED_ARG_NAME_CLS)
        if test_func_spec.num_params:
            params.append(PARAMETRIZED_ARG_NAME_FUNC)
            add_parametrize_marker(
                PARAMETRIZED_ARG_NAME_FUNC,
                num_params=test_func_spec.num_params,
                param_marker=test_func_spec.param_marker,
                with_indent=bool(test_class_spec),
            )
        sig = ", ".join(params)
        if test_class_spec:
            sig = "self, " + sig
            code += "\t"
        func_body = test_func_spec.func_body or "\tpass\n"
        code += f"def {name}({sig}):{func_body}\n"

    if isinstance(test_spec, TestFileSpec):
        for i, spec in enumerate(test_spec.test_specs, start=1):
            if isinstance(spec, TestClassSpec):
                add_class(spec)
                for j, func_spec in enumerate(spec.test_func_specs, start=1):
                    add_function(f"{TEST_NAME_BASE}{j}", func_spec, test_class_spec=spec)
            else:
                add_function(f"{TEST_NAME_BASE}{i}", spec)
    else:
        add_function(TEST_NAME_BASE, test_spec)

    ast.parse(code)
    return code


def get_num_tests(*test_specs: TestFuncSpec | TestClassSpec | TestFileSpec) -> int:
    """Calculate the total number of tests from the given test specs

    :param test_specs: One or more TestFuncSpec, TestClassSpec, or TestFileSpec objects to calculate the total number of
                       tests against
    """
    num_tests = 0
    for spec in test_specs:
        if isinstance(spec, TestFileSpec):
            num_tests += get_num_tests(*spec.test_specs)
        elif isinstance(spec, TestClassSpec):
            num_tests += get_num_tests(*spec.test_func_specs)
        else:
            if spec.test_class_spec:
                num_outer_params = spec.test_class_spec.num_params or 1
            else:
                num_outer_params = 1
            num_tests += num_outer_params * (spec.num_params or 1)
    return num_tests


def get_num_tests_to_be_selected(test_file_specs: list[TestFileSpec], n: str | None, scope: SmokeScope | None) -> int:
    """Calculate the number of expected tests to be selected from the test file specs for the given conditions

    :param test_file_specs: Test file specs
    :param n: N value given for the smoke option
    :param scope: Smoke scope value
    """

    def get_num_expected_tests_per_scope(*test_specs):
        n_ = n or str(DEFAULT_N)
        if n_.endswith("%"):
            scale = float(n_[:-1])
            return int(scale_down(get_num_tests(*test_specs), scale))
        else:
            return min([int(n_), get_num_tests(*test_specs)])

    test_class_specs = get_test_class_specs(test_file_specs) if scope in [SmokeScope.AUTO, SmokeScope.CLASS] else None
    if scope == SmokeScope.ALL:
        num_expected_tests = get_num_expected_tests_per_scope(*test_file_specs)
    elif scope == SmokeScope.DIRECTORY:
        test_specs_per_dir = {}
        for test_file_spec in test_file_specs:
            test_specs_per_dir.setdefault(test_file_spec.test_dir, []).extend(test_file_spec.test_specs)
        num_expected_tests = sum(
            get_num_expected_tests_per_scope(*test_specs) for test_specs in test_specs_per_dir.values()
        )
    elif scope == SmokeScope.FILE:
        num_expected_tests = sum(get_num_expected_tests_per_scope(*x.test_specs) for x in test_file_specs)
    elif scope == SmokeScope.AUTO:
        num_expected_tests = sum(
            [
                *(get_num_expected_tests_per_scope(x) for x in test_class_specs),
                *(
                    get_num_expected_tests_per_scope(x)
                    for x in get_test_func_specs(test_file_specs, exclude_class=True)
                ),
            ]
        )
    elif scope == SmokeScope.CLASS:
        num_expected_tests = sum(get_num_expected_tests_per_scope(x) for x in test_class_specs)
    else:
        # default = function
        num_expected_tests = sum(get_num_expected_tests_per_scope(x) for x in get_test_func_specs(test_file_specs))

    return num_expected_tests


def get_test_class_specs(test_file_specs: list[TestFileSpec]) -> list[TestClassSpec]:
    """Returns all TestClassSpec objects from test file specs

    :param test_file_specs: Test file specs
    """
    test_specs = list(chain(*[x.test_specs for x in test_file_specs]))
    return [x for x in test_specs if isinstance(x, TestClassSpec)]


def get_test_func_specs(test_file_specs: list[TestFileSpec], exclude_class: bool = False) -> list[TestFuncSpec]:
    """Returns all TestFuncSpec objects from test file specs

    :param test_file_specs: Test file specs
    :param exclude_class: Exclude test func specs defined under a test class spec
    """
    test_specs = list(chain(*[x.test_specs for x in test_file_specs]))
    if exclude_class:
        test_specs = [x for x in test_specs if not isinstance(x, TestClassSpec)]
    return list(chain(*[(x.test_func_specs if isinstance(x, TestClassSpec) else [x]) for x in test_specs]))


@contextmanager
def mock_column_width(width: int):
    """Temporarily mock the column width

    :param width: Column width
    """
    mp = pytest.MonkeyPatch()
    mp.setenv("COLUMNS", str(width))
    try:
        yield
    finally:
        mp.undo()


def patch_runpytest(f):
    """A decorator for patching pytester.runpytest() to temporarily mock the column width during a test when the -v
    option is given to ensure the stdout is captured in a standard terminal size.
    This prevents the flaky test results caused by a teser's actual terminal size
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        with mock_column_width(150) if "-v" in args else nullcontext():
            return f(*args, **kwargs)

    return wrapper



from __future__ import annotations

import os
import random
from decimal import ROUND_HALF_UP, Decimal
from functools import lru_cache
from typing import TYPE_CHECKING, Optional, Union
from uuid import UUID

import pytest

from pytest_smoke import smoke
from pytest_smoke.types import SmokeEnvVar, SmokeIniOption, SmokeOption, SmokeScope, SmokeSelectMode

if smoke.is_xdist_installed:
    from xdist import is_xdist_controller, is_xdist_worker

if TYPE_CHECKING:
    from pytest import Config, Item, Session


@lru_cache
def scale_down(value: float, percentage: float, precision: int = 0, min_value: int = 1) -> float:
    """Scales down a value with rounding

    :param value: The current value to calculate the percentage from
    :param percentage: The percentage to apply to the current value
    :param precision: The number of decimal places to round to
    :param min_value: The minimum allowed value after calculation
    """
    if percentage > 100:
        raise ValueError("The percentage must be 100 or smaller")
    val = _round_half_up(value * percentage / 100, precision)
    return max(val, min_value)


@lru_cache
def generate_group_id(item: Item, scope: str) -> Optional[str]:
    """Generate a smoke scope group ID for the item

    :param item: Collected Pytest item
    :param scope: Smoke scope
    """
    assert scope
    if item.config.hook.pytest_smoke_exclude(item=item, scope=scope):
        return

    if (group_id := item.config.hook.pytest_smoke_generate_group_id(item=item, scope=scope)) is not None:
        return group_id

    if scope not in [str(x) for x in SmokeScope]:
        raise pytest.UsageError(
            f"The logic for the custom scope '{scope}' must be implemented using the "
            f"pytest_smoke_generate_group_id hook"
        )

    if scope == SmokeScope.ALL:
        return "*"

    cls = getattr(item, "cls", None)
    if not cls and scope == SmokeScope.CLASS:
        return

    file_path = item.path
    if scope == SmokeScope.DIRECTORY:
        return str(file_path.parent)

    group_id = str(file_path)
    if scope == SmokeScope.FILE:
        return group_id

    if cls:
        group_id += f"::{cls.__name__}"
        if scope in [SmokeScope.CLASS, SmokeScope.AUTO]:
            return group_id

    # The default scope
    func_name = item.function.__name__  # type: ignore
    group_id += f"::{func_name}"
    return group_id


def sort_items(items: list[Item], session: Session, smoke_option: SmokeOption) -> list[Item]:
    """Sort collected Pytest items for the given select mode

    :param items: Collected Pytest items
    :param session: Pytest session
    :param smoke_option: Smoke option
    """
    if smoke_option.select_mode == SmokeSelectMode.FIRST:
        sorted_items = items
    elif smoke_option.select_mode == SmokeSelectMode.LAST:
        sorted_items = items[::-1]
    elif smoke_option.select_mode == SmokeSelectMode.RANDOM:
        if smoke.is_xdist_installed and (is_xdist_controller(session) or is_xdist_worker(session)):
            # Set the seed to ensure XDIST controler and workers collect the same items
            random_ = random.Random(UUID(os.environ[SmokeEnvVar.SMOKE_TEST_SESSION_UUID]).time)
        else:
            random_ = random
        sorted_items = random_.sample(items, len(items))
    else:
        sorted_items = session.config.hook.pytest_smoke_sort_by_select_mode(
            items=items.copy(), scope=smoke_option.scope, select_mode=smoke_option.select_mode
        )
        if sorted_items is None:
            raise pytest.UsageError(
                f"The custom sort logic for the select mode '{smoke_option.select_mode}' must be implemented using the "
                f"pytest_smoke_sort_by_select_mode hook"
            )
    return sorted_items


def parse_n(value: str) -> Union[int, str]:
    v = value.strip()
    try:
        if is_scale := v.endswith("%"):
            num = float(v[:-1])
        else:
            num = int(v)

        if num < 1 or is_scale and num > 100:
            raise ValueError

        if is_scale:
            return f"{num}%"
        else:
            return num
    except ValueError:
        raise pytest.UsageError(
            f"The smoke N value must be a positive number or a valid percentage. '{value}' was given."
        )


def parse_select_mode(value: str) -> str:
    if (v := value.strip()) == "":
        raise pytest.UsageError(f"Invalid select mode: '{value}'")
    return v


def parse_scope(value: str) -> str:
    if (v := value.strip()) == "":
        raise pytest.UsageError(f"Invalid scope: '{value}'")
    return v


def parse_ini_option(config: Config, option: SmokeIniOption) -> Union[str, int, bool]:
    try:
        v = config.getini(option)
        if option == SmokeIniOption.SMOKE_DEFAULT_N:
            return parse_n(v)
        elif option == SmokeIniOption.SMOKE_DEFAULT_SELECT_MODE:
            return parse_select_mode(v)
        elif option == SmokeIniOption.SMOKE_DEFAULT_SCOPE:
            return parse_scope(v)
        else:
            return v
    except ValueError as e:
        raise pytest.UsageError(f"{option}: {e}")


def _round_half_up(x: float, precision: int) -> float:
    return float(Decimal(str(x)).quantize(Decimal("10") ** -precision, rounding=ROUND_HALF_UP))



import sys
from enum import Enum

import pytest

if sys.version_info < (3, 11):

    class StrEnum(str, Enum):
        def _generate_next_value_(name, start, count, last_values) -> str:
            return name.lower()

        def __str__(self) -> str:
            return str(self.value)
else:
    from enum import StrEnum  # noqa: F401


if pytest.version_tuple < (7, 4):
    from collections.abc import Mapping
    from typing import NamedTuple, Union

    class TestShortLogReport(NamedTuple):
        category: str
        letter: str
        word: Union[str, tuple[str, Mapping[str, bool]]]
else:
    from pytest import TestShortLogReport  # noqa: F401



import re

import pytest
from pytest import ExitCode, Pytester

from pytest_smoke.utils import scale_down
from tests.helper import TEST_NAME_BASE, TestFuncSpec, generate_test_code


@pytest.mark.parametrize("with_hook", [True, False])
def test_smoke_hook_pytest_smoke_generate_group_id(pytester: Pytester, with_hook: bool):
    """Test pytest_smoke_generate_group_id hook and custome scopes, with/without hook definition"""
    custom_scope = "my-scope"
    num_tests = 10
    smoke_n = 2
    num_expected_selected_tests = smoke_n * 2
    assert num_expected_selected_tests < num_tests
    pytester.makepyfile(generate_test_code(TestFuncSpec(num_params=num_tests)))
    if with_hook:
        pytester.makeconftest(f"""
        import pytest
        def pytest_smoke_generate_group_id(item, scope):
            if scope == "{custom_scope}":
                # group tests by odd/even index
                return item.session.items.index(item) % 2
        """)
    result = pytester.runpytest("--smoke", str(smoke_n), "--smoke-scope", custom_scope)
    if not with_hook:
        assert result.ret == ExitCode.USAGE_ERROR
        result.stderr.re_match_lines(
            [
                f"ERROR: The logic for the custom scope '{custom_scope}' must be implemented using the "
                f"pytest_smoke_generate_group_id hook"
            ]
        )
    else:
        assert result.ret == ExitCode.OK
        result.assert_outcomes(passed=num_expected_selected_tests, deselected=num_tests - num_expected_selected_tests)


def test_smoke_hook_pytest_smoke_include(pytester: Pytester):
    """Test pytest_smoke_include hook"""
    num_tests = 10
    n = 2
    num_include = 2
    num_expected_selected_tests = n + num_include
    assert num_expected_selected_tests < num_tests
    param_marker = lambda p: "smoke" if p < num_include else None  # noqa
    pytester.makepyfile(generate_test_code(TestFuncSpec(num_params=num_tests, param_marker=param_marker)))
    pytester.makeconftest("""
    import pytest
    def pytest_smoke_include(item, scope):
        # Include tests with the smoke mark as additional tests
        return item.get_closest_marker("smoke")
    """)
    result = pytester.runpytest("--smoke", str(n))
    assert result.ret == ExitCode.OK
    result.assert_outcomes(passed=num_expected_selected_tests, deselected=num_tests - num_expected_selected_tests)


@pytest.mark.parametrize("n", ["2", "10", "100", "20%", "100%"])
def test_smoke_hook_pytest_smoke_exclude(pytester: Pytester, n: str):
    """Test pytest_smoke_exclude hook"""
    num_tests = 20
    divisor = 3
    num_selectable_tests = int(sum([x % divisor not in [1, 2] for x in range(num_tests)]))
    is_scale = n.endswith("%")
    if is_scale:
        num_expected_selected_tests = scale_down(num_selectable_tests, float(n[:-1]))
    else:
        num_expected_selected_tests = min([int(n), num_selectable_tests])

    def param_marker(i):
        reminder = i % divisor
        if reminder == 1:
            return "skip"
        elif reminder == 2:
            return "xfail"

    pytester.makepyfile(generate_test_code(TestFuncSpec(num_params=num_tests, param_marker=param_marker)))
    pytester.makeconftest("""
    import pytest
    def pytest_smoke_exclude(item, scope):
        # Exclude tests with skip/xfail marks
        return any(item.get_closest_marker(m) for m in ["skip", "xfail"])
    """)
    result = pytester.runpytest("--smoke", str(n))
    assert result.ret == ExitCode.OK
    result.assert_outcomes(passed=num_expected_selected_tests, deselected=num_tests - num_expected_selected_tests)


@pytest.mark.parametrize("with_hook", [True, False])
def test_smoke_hook_pytest_smoke_sort_by_select_mode(pytester: Pytester, with_hook: bool):
    """Test custom select mode using the pytest_smoke_sort_by_select_mode hook, with/without hook
    definition
    """
    custom_select_mode = "my-select-mode"
    num_tests = 10
    smoke_n = 2
    pytester.makepyfile(generate_test_code(TestFuncSpec(num_params=num_tests)))
    if with_hook:
        pytester.makeconftest("""
        import pytest
        def pytest_smoke_sort_by_select_mode(items):
            # sort tests by odd/even index
            return sorted(items, key=lambda x: items.index(x) % 2)
        """)
    result = pytester.runpytest("--smoke", str(smoke_n), "--smoke-select-mode", custom_select_mode, "-v")
    if not with_hook:
        assert result.ret == ExitCode.USAGE_ERROR
        result.stderr.re_match_lines(
            [
                f"ERROR: The custom sort logic for the select mode '{custom_select_mode}' must be implemented "
                f"using the pytest_smoke_sort_by_select_mode hook"
            ]
        )
    else:
        assert result.ret == ExitCode.OK
        result.assert_outcomes(passed=smoke_n, deselected=num_tests - smoke_n)
        matched_test_nums = re.findall(rf"test_.+\.py::{TEST_NAME_BASE}\[(\d+)\]", str(result.stdout))
        assert len(matched_test_nums) == smoke_n
        assert [int(n) for n in matched_test_nums] == sorted([x for x in range(num_tests)], key=lambda x: x % 2)[
            :smoke_n
        ]



from pytest import Config, Item, Session, hookimpl

from pytest_smoke import smoke
from pytest_smoke.types import SmokeIniOption, SmokeOption
from pytest_smoke.utils import generate_group_id, parse_ini_option

if smoke.is_xdist_installed:
    from xdist import is_xdist_controller
    from xdist.scheduler import LoadScopeScheduling

    class PytestSmokeXdist:
        """A plugin that extends pytest-smoke to seamlesslly support pytest-xdist

        This plugin will be dynamically registered when the -n/--numprocesses option is given
        """

        name = "smoke-xdist"

        def __init__(self):
            self._nodes = None

        @hookimpl(tryfirst=True)
        def pytest_collection(self, session: Session):
            if is_xdist_controller(session):
                self._nodes = {item.nodeid: item for item in session.perform_collect()}
                return True

        def pytest_xdist_make_scheduler(self, config: Config, log):
            """Replace the pytest-xdist default scheduler (load) with our custom scheduler (smoke scope) when the
            following conditions match:
            - The INI option value is set to true
            - No dist option (--dist or -d) is explicitly given
            """
            if (
                config.known_args_namespace.dist == "no"
                and not config.known_args_namespace.distload
                and parse_ini_option(config, SmokeIniOption.SMOKE_DEFAULT_XDIST_DIST_BY_SCOPE)
            ):
                return SmokeScopeScheduling(config, log, nodes=self._nodes)

    class SmokeScopeScheduling(LoadScopeScheduling):
        """A custom pytest-xdist scheduler that distributes workloads by smoke scope groups"""

        def __init__(self, config: Config, log, *, nodes: dict[str, Item]):
            super().__init__(config, log)
            self.smoke_option = SmokeOption(config)
            self._nodes = nodes

        def _split_scope(self, nodeid: str) -> str:
            item = self._nodes[nodeid]

            return generate_group_id(item, self.smoke_option.scope) or super()._split_scope(nodeid)



from __future__ import annotations

import os
from collections import Counter
from typing import TYPE_CHECKING, Optional, Union
from uuid import uuid4

import pytest
from pytest import StashKey

from pytest_smoke import smoke
from pytest_smoke.compat import TestShortLogReport
from pytest_smoke.types import (
    SmokeCounter,
    SmokeDefaultN,
    SmokeEnvVar,
    SmokeIniOption,
    SmokeMarker,
    SmokeOption,
    SmokeScope,
    SmokeSelectMode,
)
from pytest_smoke.utils import (
    generate_group_id,
    parse_ini_option,
    parse_n,
    parse_scope,
    parse_select_mode,
    scale_down,
    sort_items,
)

if smoke.is_xdist_installed:
    from xdist import is_xdist_worker

    from pytest_smoke.extensions.xdist import PytestSmokeXdist

if TYPE_CHECKING:
    from pytest import Config, Item, Parser, PytestPluginManager, Session, StashKey, TestReport


STASH_KEY_SMOKE_COUNTER = StashKey[SmokeCounter]()
STASH_KEY_SMOKE_IS_CIRITICAL = StashKey[bool]()
STASH_KEY_SMOKE_IS_MUSTPASS = StashKey[bool]()
STASH_KEY_SMOKE_SHOULD_SKIP_RESET = StashKey[bool]()
DEFAULT_N = SmokeDefaultN(1)


@pytest.hookimpl(trylast=True)
def pytest_addhooks(pluginmanager: PytestPluginManager):
    from pytest_smoke import hooks

    pluginmanager.add_hookspecs(hooks)


def pytest_addoption(parser: Parser):
    group = parser.getgroup("smoke", description="Smoke testing")
    group.addoption(
        "--smoke",
        dest="smoke",
        metavar="N",
        const=DEFAULT_N,
        type=parse_n,
        nargs="?",
        default=False,
        help="Run only N tests from each test function or specified scope.\n"
        "N can be a number (e.g. 5) or a percentage (e.g. 10%%).\n"
        "If not provided, the default value is 1.",
    )
    group.addoption(
        "--smoke-scope",
        dest="smoke_scope",
        metavar="SCOPE",
        type=parse_scope,
        help=(
            "Specify the scope at which the value of N from the above options is applied.\n"
            "The plugin provides the following predefined scopes, as well as custom user-defined scopes via a hook:\n"
            f"- {SmokeScope.FUNCTION}: Applies to each test function (default)\n"
            f"- {SmokeScope.CLASS}: Applies to each test class\n"
            f"- {SmokeScope.AUTO}: Applies {SmokeScope.FUNCTION} scope for test functions, "
            f"{SmokeScope.CLASS} scope for test methods\n"
            f"- {SmokeScope.FILE}: Applies to each test file\n"
            f"- {SmokeScope.DIRECTORY}: Applies to each test directory\n"
            f"- {SmokeScope.ALL}: Applies to the entire test suite"
        ),
    )
    group.addoption(
        "--smoke-select-mode",
        dest="smoke_select_mode",
        metavar="MODE",
        type=parse_select_mode,
        help=(
            "Specify the mode for selecting tests from each scope.\n"
            "The plugin provides the following predefined values, as well as custom user-defined values via a hook:\n"
            f"- {SmokeSelectMode.FIRST}: The first N tests (default)\n"
            f"- {SmokeSelectMode.LAST}: The last N tests\n"
            f"- {SmokeSelectMode.RANDOM}: N randomly selected tests"
        ),
    )

    parser.addini(
        SmokeIniOption.SMOKE_DEFAULT_N,
        type="string",
        default=str(DEFAULT_N),
        help="[pytest-smoke] Override the plugin default value for smoke N",
    )
    parser.addini(
        SmokeIniOption.SMOKE_DEFAULT_SCOPE,
        type="string",
        default=SmokeScope.FUNCTION,
        help="[pytest-smoke] Override the plugin default value for smoke scope",
    )
    parser.addini(
        SmokeIniOption.SMOKE_DEFAULT_SELECT_MODE,
        type="string",
        default=SmokeSelectMode.FIRST,
        help="[pytest-smoke] Override the plugin default value for smoke select mode",
    )
    parser.addini(
        SmokeIniOption.SMOKE_DEFAULT_XDIST_DIST_BY_SCOPE,
        type="bool",
        default=False,
        help="[pytest-smoke] When using the pytest-xdist plugin for parallel testing, replace the default scheduler "
        "with a custom distribution algorithm that distributes tests based on the smoke scope",
    )
    parser.addini(
        SmokeIniOption.SMOKE_MARKED_TESTS_AS_CRITICAL,
        type="bool",
        default=False,
        help="[pytest-smoke] Treat tests marked with @pytest.mark.smoke as 'critical' smoke tests",
    )


@pytest.hookimpl(tryfirst=True)
def pytest_configure(config: Config):
    if not config.option.smoke and (config.option.smoke_scope or config.option.smoke_select_mode):
        raise pytest.UsageError("The --smoke option is requierd to use the pytest-smoke functionality")

    config.addinivalue_line(
        "markers",
        "smoke(*, mustpass=False, runif=True): [pytest-smoke] When running smoke tests using the pytest-smoke plugin, "
        "and the feature is explicitly enabled via an INI option, the marked test is considered a 'critical' smoke "
        "test. Additionally, if the optional mustpass keyword argument is set to True, the test is considered a "
        "'must-pass' critical smoke test. Critical smoke tests with runif=True are automatically included and executed "
        "first, before regular smoke tests. If any 'must-pass' test fails, all subsequent regular smoke tests will be "
        "skipped.\n"
        "Note: The marker will have no effect on the plugin until the feature has been enabled",
    )

    if smoke.is_xdist_installed:
        if config.pluginmanager.has_plugin("xdist"):
            # Register the smoke-xdist plugin if -n/--numprocesses option is given.
            if config.getoption("numprocesses", default=None):
                config.pluginmanager.register(PytestSmokeXdist(), name=PytestSmokeXdist.name)
        else:
            smoke.is_xdist_installed = False


@pytest.hookimpl(wrapper=True, tryfirst=True)
def pytest_sessionstart(session: Session):
    if not smoke.is_xdist_installed or not is_xdist_worker(session):
        os.environ[SmokeEnvVar.SMOKE_TEST_SESSION_UUID] = str(uuid4())
    return (yield)


@pytest.hookimpl(wrapper=True, trylast=True)
def pytest_collection_modifyitems(session: Session, config: Config, items: list[Item]):
    try:
        return (yield)
    finally:
        if not items:
            return

        opt = SmokeOption(config)
        if opt.n:
            selected_items_regular = []
            selected_items_critical = []
            deselected_items = []
            smoke_groups_reached_threshold = set()
            counter = SmokeCounter(
                collected=Counter(filter(None, (generate_group_id(item, opt.scope) for item in items)))
            )
            session.stash[STASH_KEY_SMOKE_COUNTER] = counter
            enable_critical_tests = parse_ini_option(config, SmokeIniOption.SMOKE_MARKED_TESTS_AS_CRITICAL)

            for item in sort_items(items, session, opt):
                group_id = generate_group_id(item, opt.scope)
                if group_id is None:
                    deselected_items.append(item)
                    continue

                # Tests that match the below conditions will not be counted towards the calculation of N
                if enable_critical_tests and (smoke_marker := SmokeMarker.from_item(item)):
                    if smoke_marker.runif:
                        selected_items_critical.append(item)
                        if smoke_marker.mustpass:
                            counter.mustpass.selected.add(item)
                        item.stash[STASH_KEY_SMOKE_IS_CIRITICAL] = True
                        item.stash[STASH_KEY_SMOKE_IS_MUSTPASS] = smoke_marker.mustpass
                    else:
                        deselected_items.append(item)
                    continue
                elif config.hook.pytest_smoke_include(item=item, scope=opt.scope):
                    selected_items_regular.append(item)
                    continue

                if group_id in smoke_groups_reached_threshold:
                    deselected_items.append(item)
                    continue

                threshold = scale_down(counter.collected[group_id], float(opt.n[:-1])) if opt.is_scale else opt.n
                if counter.selected[group_id] < threshold:
                    counter.selected.update([group_id])
                    selected_items_regular.append(item)
                else:
                    smoke_groups_reached_threshold.add(group_id)
                    deselected_items.append(item)

            assert len(items) == len(selected_items_critical + selected_items_regular + deselected_items)
            if selected_items_critical or deselected_items:
                if deselected_items:
                    config.hook.pytest_deselected(items=deselected_items)

                if opt.select_mode != SmokeSelectMode.FIRST:
                    # retain the original test order
                    for smoke_items in (selected_items_critical, selected_items_regular):
                        if smoke_items:
                            smoke_items.sort(key=lambda x: items.index(x))

                items.clear()
                items.extend(selected_items_critical + selected_items_regular)


@pytest.hookimpl(wrapper=True)
def pytest_runtest_protocol(item: Item, nextitem: Optional[Item]):
    try:
        return (yield)
    finally:
        if nextitem and item.stash.get(STASH_KEY_SMOKE_IS_CIRITICAL, False):
            counter = item.session.stash[STASH_KEY_SMOKE_COUNTER].mustpass
            if counter.failed and not nextitem.stash.get(STASH_KEY_SMOKE_IS_CIRITICAL, False):
                # At least one must-pass test failed, and this is the last critical test.
                # Set the flag to skip all subsequent regular tests
                item.session.stash[STASH_KEY_SMOKE_SHOULD_SKIP_RESET] = True


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_setup(item: Item):
    if item.session.stash.get(STASH_KEY_SMOKE_SHOULD_SKIP_RESET, False):
        counter = item.session.stash[STASH_KEY_SMOKE_COUNTER].mustpass
        num_selected = len(counter.selected)
        num_failed = len(counter.failed)
        pytest.skip(reason=f"{num_failed}/{num_selected} must-pass smoke test{'s' if num_failed > 1 else ''} failed")


@pytest.hookimpl(wrapper=True)
def pytest_runtest_makereport(item: Item):
    report: TestReport = yield
    if item.stash.get(STASH_KEY_SMOKE_IS_MUSTPASS, False):
        setattr(report, "_is_smoke_must_pass", True)
        if report.failed:
            item.session.stash[STASH_KEY_SMOKE_COUNTER].mustpass.failed.add(item)
    return report


@pytest.hookimpl(wrapper=True, trylast=True)
def pytest_report_teststatus(report: TestReport):
    status: Union[tuple, TestShortLogReport] = yield
    if not isinstance(status, TestShortLogReport):
        status = TestShortLogReport(*status)
    if status.word and getattr(report, "_is_smoke_must_pass", False):
        annot = " (must-pass)"
        if isinstance(status.word, str):
            status = status._replace(word=status.word + annot)
        elif isinstance(status.word, tuple):
            status = status._replace(word=([status.word[0] + annot, *status.word[1:]]))
    return status



import re
from typing import Optional

import pytest
from pytest import ExitCode, Pytester

from pytest_smoke import smoke
from pytest_smoke.types import SmokeIniOption, SmokeScope, SmokeSelectMode
from tests.helper import TEST_NAME_BASE, TestFileSpec, TestFuncSpec, generate_test_code, requires_xdist

if smoke.is_xdist_installed:
    from xdist.scheduler import LoadScheduling

    from pytest_smoke.extensions.xdist import SmokeScopeScheduling


SMOKE_INI_OPTIONS = [
    SmokeIniOption.SMOKE_DEFAULT_N,
    SmokeIniOption.SMOKE_DEFAULT_SCOPE,
    pytest.param(SmokeIniOption.SMOKE_DEFAULT_XDIST_DIST_BY_SCOPE, marks=requires_xdist),
]


def test_smoke_ini_option_smoke_default_n(pytester: Pytester):
    """Test smoke_default_n INI option"""
    num_tests = 10
    default_n = 3
    pytester.makepyfile(generate_test_code(TestFuncSpec(num_params=num_tests)))
    pytester.makeini(f"""
    [pytest]
    {SmokeIniOption.SMOKE_DEFAULT_N} = {default_n}
    """)
    result = pytester.runpytest("--smoke")
    assert result.ret == ExitCode.OK
    result.assert_outcomes(passed=default_n, deselected=num_tests - default_n)


def test_smoke_ini_option_smoke_default_scope(pytester: Pytester):
    """Test smoke_default_scope INI option"""
    num_tests_1 = 5
    num_tests_2 = 10
    pytester.makepyfile(
        generate_test_code(TestFileSpec([TestFuncSpec(num_params=num_tests_1), TestFuncSpec(num_params=num_tests_2)]))
    )
    pytester.makeini(f"""
    [pytest]
    {SmokeIniOption.SMOKE_DEFAULT_SCOPE} = {SmokeScope.FILE}
    """)
    result = pytester.runpytest("--smoke")
    assert result.ret == ExitCode.OK
    result.assert_outcomes(passed=1, deselected=num_tests_1 + num_tests_2 - 1)


def test_smoke_ini_option_smoke_default_select_mode(pytester: Pytester):
    """Test smoke_default_select_mode INI option"""
    smoke_n = 3
    num_tests = 10
    pytester.makepyfile(generate_test_code(TestFuncSpec(num_params=num_tests)))
    pytester.makeini(f"""
    [pytest]
    {SmokeIniOption.SMOKE_DEFAULT_SELECT_MODE} = {SmokeSelectMode.LAST}
    """)
    result = pytester.runpytest("--smoke", str(smoke_n), "--co", "-q")
    assert result.ret == ExitCode.OK
    result.assert_outcomes(deselected=num_tests - smoke_n)

    matched_test_nums = re.findall(rf"test_.+\.py::{TEST_NAME_BASE}\[(\d+)\]", str(result.stdout))
    assert len(matched_test_nums) == smoke_n
    assert [int(n) for n in matched_test_nums] == list(range(num_tests))[-smoke_n:]


@requires_xdist
@pytest.mark.parametrize("dist_option", [None, "--dist=load", "-d"])
@pytest.mark.parametrize("value", ["true", "false", None])
def test_smoke_ini_option_smoke_default_xdist_dist_by_scope(
    pytester: Pytester, value: Optional[str], dist_option: Optional[str]
):
    """Test smoke_default_xdist_dist_by_scope INI option.

    The plugin should extend the pytest-xdist to use the custom scheduler when the INI option value is true and
    when no dist option (--dist or -d) is explicitly given
    """
    from xdist import __version__ as xdist_ver

    num_tests_1 = 100
    num_tests_2 = 100
    test_file_spec = TestFileSpec([TestFuncSpec(num_params=num_tests_1), TestFuncSpec(num_params=num_tests_2)])
    num_test_func = len(test_file_spec.test_specs)
    smoke_n = 99
    pytester.makepyfile(test_xdist=generate_test_code(test_file_spec))
    if value:
        pytester.makeini(f"""
        [pytest]
        {SmokeIniOption.SMOKE_DEFAULT_XDIST_DIST_BY_SCOPE} = {value}
        """)
    args = ["--smoke", str(smoke_n), "-v", "-n", "2"]
    if dist_option:
        args.append(dist_option)
        if tuple(map(int, xdist_ver.split("."))) >= (3, 2, 0):
            # Sometimes our github test workflow on windows fails as tests don't get distributed evenly.
            # Add --maxschedchunk=1 as a workaround
            args.extend(["--maxschedchunk", "1"])

    result = pytester.runpytest(*args)
    assert result.ret == ExitCode.OK
    expected_scheduler = SmokeScopeScheduling if dist_option is None and value == "true" else LoadScheduling
    result.stdout.re_match_lines(f"scheduling tests via {expected_scheduler.__name__}")
    num_passed = num_test_func * smoke_n
    result.assert_outcomes(passed=num_passed, deselected=num_tests_1 + num_tests_2 - num_passed)

    test_ids_worker1 = re.findall(rf"\[gw0\].+PASSED (test_xdist\.py::{TEST_NAME_BASE}\d)\[\d+\]", str(result.stdout))
    test_ids_worker2 = re.findall(rf"\[gw1\].+PASSED (test_xdist\.py::{TEST_NAME_BASE}\d)\[\d+\]", str(result.stdout))
    assert test_ids_worker1 and test_ids_worker2
    assert len(test_ids_worker1) + len(test_ids_worker2) == num_passed
    for test_ids in (test_ids_worker1, test_ids_worker2):
        if expected_scheduler == SmokeScopeScheduling:
            # Since the default smoke scope is function, each worker processes tests only for single test function
            assert len(test_ids) == smoke_n
            assert len(set(test_ids)) == 1
        else:
            # Tests should be distributed from both test functions
            assert len(set(test_ids)) == num_test_func


@pytest.mark.parametrize("value", [None, "true", "false"])
def test_smoke_ini_option_smoke_marked_tests_as_critical(pytester: Pytester, value: Optional[str]):
    """Test smoke_marked_tests_as_critical INI option"""
    is_enabled = value == "true"
    num_tests_1 = 5
    num_tests_2 = 5
    test_file_spec = TestFileSpec(
        [
            TestFuncSpec(num_params=num_tests_1),
            TestFuncSpec(
                num_params=num_tests_2,
                # Apply @pytest.mark.smoke to all tests
                param_marker=lambda x: "smoke",
            ),
        ]
    )
    pytester.makepyfile(generate_test_code(test_file_spec))
    if value:
        pytester.makeini(f"""
        [pytest]
        {SmokeIniOption.SMOKE_MARKED_TESTS_AS_CRITICAL} = {value}
        """)
    result = pytester.runpytest("--smoke", "-v")
    assert result.ret == ExitCode.OK
    test_nums = re.findall(rf"test_.+\.py::{TEST_NAME_BASE}(\d).+", str(result.stdout))
    if is_enabled:
        assert sorted(test_nums, reverse=True) == test_nums
        num_passes = 1 + num_tests_2
    else:
        assert sorted(test_nums) == test_nums
        num_passes = len(test_file_spec.test_specs)
    result.assert_outcomes(passed=num_passes, deselected=num_tests_1 + num_tests_2 - num_passes)


@pytest.mark.parametrize("ini_option", SMOKE_INI_OPTIONS)
@pytest.mark.parametrize("value", ["foo", ""])
def test_smoke_ini_option_with_invalid_value(pytester: Pytester, ini_option: str, value: str):
    """Test INI options with an invalid value are handled as a usage error"""
    pytester.makepyfile(generate_test_code(TestFuncSpec()))
    pytester.makeini(f"""
    [pytest]
    {ini_option} = {value}
    """)
    args = ["--smoke"]
    if ini_option == SmokeIniOption.SMOKE_DEFAULT_XDIST_DIST_BY_SCOPE:
        args.extend(["-n", "2"])
    result = pytester.runpytest(*args)
    assert result.ret == ExitCode.USAGE_ERROR
